

def rgb_to_ycbcr(image: "torch.Tensor") -> "torch.Tensor":
    import torch
    r"""Convert an RGB image to YCbCr.

    Args:
        image (torch.Tensor): RGB Image to be converted to YCbCr with shape :math:`(*, 3, H, W)`.

    Returns:
        torch.Tensor: YCbCr version of the image with shape :math:`(*, 3, H, W)`.

    Examples:
        >>> input = torch.rand(2, 3, 4, 5)
        >>> output = rgb_to_ycbcr(input)  # 2x3x4x5
    """
    if not isinstance(image, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(image)))

    if len(image.shape) < 3 or image.shape[-3] != 3:
        raise ValueError("Input size must have a shape of (*, 3, H, W). Got {}"
                         .format(image.shape))

    r: torch.Tensor = image[..., 0, :, :]
    g: torch.Tensor = image[..., 1, :, :]
    b: torch.Tensor = image[..., 2, :, :]

    delta: float = 0.5
    y: torch.Tensor = 0.299 * r + 0.587 * g + 0.114 * b
    cb: torch.Tensor = (b - y) * 0.564 + delta
    cr: torch.Tensor = (r - y) * 0.713 + delta
    return torch.stack([y, cb, cr], -3)


def ycbcr_to_rgb(image: "torch.Tensor") -> "torch.Tensor":
    import torch
    r"""Convert an YCbCr image to RGB.

    The image data is assumed to be in the range of (0, 1).

    Args:
        image (torch.Tensor): YCbCr Image to be converted to RGB with shape :math:`(*, 3, H, W)`.

    Returns:
        torch.Tensor: RGB version of the image with shape :math:`(*, 3, H, W)`.

    Examples:
        >>> input = torch.rand(2, 3, 4, 5)
        >>> output = ycbcr_to_rgb(input)  # 2x3x4x5
    """
    if not isinstance(image, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(image)))

    if len(image.shape) < 3 or image.shape[-3] != 3:
        raise ValueError("Input size must have a shape of (*, 3, H, W). Got {}"
                         .format(image.shape))

    y: torch.Tensor = image[..., 0, :, :]
    cb: torch.Tensor = image[..., 1, :, :]
    cr: torch.Tensor = image[..., 2, :, :]

    delta: float = 0.5
    cb_shifted: torch.Tensor = cb - delta
    cr_shifted: torch.Tensor = cr - delta

    r: torch.Tensor = y + 1.403 * cr_shifted
    g: torch.Tensor = y - 0.714 * cr_shifted - 0.344 * cb_shifted
    b: torch.Tensor = y + 1.773 * cb_shifted
    return torch.stack([r, g, b], -3)



import io
import itertools
from functools import partial
from typing import List, Any

import numpy as np
from loguru import logger
from matplotlib import pyplot as plt

from omni_converter.solver.rules import RuleEdge


def to_nrm_quiver_auto(nrm: "AutoData",auto_func) -> "AutoData":
    return to_nrm_quiver(nrm.to("numpy,float32,HWC,XYZ,-1_1"),auto_func)


def to_nrm_quiver(nrm: np.ndarray,auto_func) -> "AutoData":
    # fmt: numpy,float32,HWC,XYZ,-1_1
    # converting an image to nrm_quiver destructs semantic consistency. should I avoid adding this to rule?
    # I guess not.
    # we need 10 for 256 px so
    np_nrm = nrm  # == ("numpy,float32,HWC,XYZ,-1_1")
    h, w, c = np_nrm.shape
    figw = w / 256 * 10
    figh = h / 256 * 10
    logger.info(f"using fig size:{(figw, figh)}")
    fig, ax = plt.subplots(figsize=(figw, figh))
    quiver_sample_n = int(min(w, h) / 256 * 50)

    x, y = np.meshgrid(np.linspace(0, w - 1, quiver_sample_n), np.linspace(0, h - 1, quiver_sample_n))
    sampled = np_nrm[-y.astype(int), x.astype(int)]  # (n,n,3)
    quiver = ax.quiver(x, y, sampled[:, :, 0], sampled[:, :, 1])
    ax.tick_params(length=5)
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    img = plt.imread(buf)
    return auto_func("numpy,float32,HWC,RGBA,0_1",img)


def swap_channel_order(img: np.ndarray, src_idx, dst_idx):
    assert len(img.shape) == 4, "only support BHWC type of ndarray image"
    copied = np.copy(img)
    src = copied[:, :, :, src_idx]
    copied[:, :, :, src_idx] = copied[:, :, :, dst_idx]
    copied[:, :, :, dst_idx] = src
    return src


import re

CH_MATCHER = re.compile("[A-Z][a-z]*")


def extract_channels(ch_repr: str):
    return CH_MATCHER.findall(ch_repr)


def swap_element(items: list, i, j):
    items = items.copy()
    tmp = items[i]
    items[i] = items[j]
    items[j] = tmp
    return items


def rule_to_swap_channel_order(state):
    from pampy import match
    def _swap_rules(s: dict):
        channels = extract_channels(s["ch_rpr"])
        axis = list(range(len(channels)))
        res = []
        for src_i, dst_i in itertools.combinations(axis, 2):
            new_chs = swap_element(channels, src_i, dst_i)
            new_state = s.copy()
            new_state["ch_rpr"] = "".join(new_chs)
            res.append(RuleEdge(
                converter=partial(swap_channel_order, src_idx=src_i, dst_idx=dst_i),
                new_state=new_state,
                cost=1,
                name=f"swap channel {channels[src_i]} to {channels[dst_i]}"
            ))
        return res

    match(state,
          {"type": "numpy", "arrange": "BHWC"}, _swap_rules,
          Any, lambda state: None
          )


def rule_to_nrm_quiver(state) -> List[RuleEdge]:
    if state == "numpy,float32,HWC,XYZ,-1_1":
        return [RuleEdge(
            to_nrm_quiver,
            "nrm_quiver",
            1,
            "normal map to quiver plot"
        )]
